//
//  MINTLJumpbar.swift
//  TreeOutlinerSwift
//
//  Created by narita on 2017/10/17.
//  Copyright © 2017年 narita. All rights reserved.
//

import Cocoa


@objc public protocol MINTLJumpbarDelegate : NSPathControlDelegate
{
    // 深さをdelegateに聞く。aa>bb>ccは深さ3。aaは深さ1。深さ0はありえない。
    @objc  func titleDepth(_ jumpbar: MINTLJumpbar) -> Int

    // aa>bb>ccで、index=0の時は、"aa"を返す。index=2の時は"cc"を返す。0ベース。深さ以下。
    @objc  func jumpbar(_ jumpbar: MINTLJumpbar, titleIndex: Int) -> String

    // 普通にクリックされた場合に呼び出される
    @objc optional func jumpbar(_ jumpbar: MINTLJumpbar, click index: Int)
    
    // 長クリックされた時に表示するメニューを返す
    @objc optional func jumpbar(_ jumpbar: MINTLJumpbar, menuFrom index: Int) -> NSMenu?
}

@objc public class MINTLJumpbar: NSPathControl
{
    // MARK: - class methods
    
    // MARK: - init methods
    
    // MARK: - dealloc
    
    // MARK: - NSCopying, hash, isEqual:
    
    // MARK: - NSCoding methods
    
    // MARK: - restorableState methods
    
    // MARK: - life cycle methods
    
    // MARK: - action methods
    
    @objc func reload()
    {
        if let theDelegate = self.delegate as! MINTLJumpbarDelegate!
        {
            let theDepth: Int = theDelegate.titleDepth(self)
            
            self.pathItems = (0..<theDepth).map
            {
                let theItem = NSPathControlItem.init()

                theItem.title = theDelegate.jumpbar(self, titleIndex: $0)
                
                return theItem
            }
        }
        
        self.needUpdateUI = false
    }
    
    @objc func setNeedsUpdateUI()
    {
        self.needUpdateUI = true
        self.setNeedsDisplay()
    }
    
    // MARK: - event handling methods
    
    override public func mouseDown(with event: NSEvent)
    {
        // leftMouseDraggedイベントも取得する理由は、
        // ユーザーがクリック後にマウスを動かすのはメニュー表示をしたい時の動作だからです。
        let theTmpEvent = NSApp.nextEvent(matching: [.leftMouseUp, .leftMouseDragged],
                                             until: self.doubleClickThreshold(),
                                            inMode: .eventTrackingRunLoopMode,
                                           dequeue: true)
        
        let (theClicledIndex, theRect) = self.clickedIndexAndRect(from:event)
        
        if theTmpEvent?.type == .leftMouseUp
        {
            if let theDelegate = self.delegate as! MINTLJumpbarDelegate!
            {
                if theDelegate.responds(to: #selector(MINTLJumpbarDelegate.jumpbar(_:click:)))
                {
                    theDelegate.jumpbar!(self, click: theClicledIndex)
                }
            }
        }
        else
        {
            // メニューを表示する。メニューが選ばれると、MenuItemに設定されたActionが実行される。
            // なので、コールバックを設定するのではなく、delegate側で都合の良いMenuを作成する事。
            
            if let theDelegate = self.delegate as! MINTLJumpbarDelegate!
            {
                if theDelegate.responds(to: #selector(MINTLJumpbarDelegate.jumpbar(_:menuFrom:)))
                {
                    let theMenu = theDelegate.jumpbar!(self, menuFrom: theClicledIndex)
                    let theSelectedMenuItem: NSMenuItem?
                    
                    // 選択されている最初のMenuItemを探し出す。
                    theSelectedMenuItem = theMenu?.items.filter{ $0.state == .on  }.first
                    
                    theMenu?.popUp(positioning: theSelectedMenuItem,
                                   at: NSMakePoint(NSMinX(theRect) - 17, NSMinY(theRect) + 2),
                                   in: self)
                }
            }
        }
    }
    
    // NSEventのカテゴリへ移動すること
    // マウスクリックの時間の間隔を返す
    func doubleClickThreshold() -> Date
    {
        // see http://www.cocoabuilder.com/archive/message/cocoa/2003/10/22/90316
        var theDelayTime = UserDefaults.standard.float(forKey: "com.apple.mouse.doubleClickThreshold")
        
        theDelayTime = min(max(theDelayTime, 0.2), 0.75)

        return Date.init(timeIntervalSinceNow: TimeInterval(theDelayTime))
    }
    
    func clickedIndexAndRect(from event: NSEvent) -> (Int, NSRect)
    {
        // self.clickedPathItemはmouseUpが呼び出された時に確定するらしいので、
        // NSEventの値から自前でIndexを算出する
        
        let thePathCell = self.cell as! NSPathCell
        let theClickedPoint = self.convert(event.locationInWindow, from: nil)
        let theClickedComponent = thePathCell.pathComponentCell(at: theClickedPoint,
                                                                withFrame: self.bounds,
                                                                in: self)
        let theClickedIndex = thePathCell.pathComponentCells.index(of: theClickedComponent!)!

        // 面倒なので、ここで矩形も取ってくる。theClickedComponentがnilの場合もあるはずだけどもう眠い
        let theRect = thePathCell.rect(of: theClickedComponent!, withFrame: self.bounds, in: self)
        
        return (theClickedIndex, theRect)
    }
    
    
    override public func menu(for event: NSEvent) -> NSMenu?
    {
        if event.type != .leftMouseDown
        {
            return nil
        }
        
        return super.menu(for: event)
    }
    
    // MARK: - drawing methods
    
    override public func draw(_ dirtyRect: NSRect)
    {
        if self.needUpdateUI
        {
            reload()
        }
        
        super.draw(dirtyRect)
    }

    
    // MARK: - delegate/datasource methods
    
    // MARK: - accessor methods (in pairs)
    
    var needUpdateUI: Bool = false
    
    // MARK: - Utility methods
}
